from typing import Generic, Iterable, TypeVar

from rainbowadn.core import *
from .keyed import *

__all__ = ('KeyValue', 'KeyValueFactory',)

KVKeyType = TypeVar('KVKeyType')
KVValueType = TypeVar('KVValueType')


class KeyValue(Keyed[KVKeyType], Generic[KVKeyType, KVValueType]):
    def __init__(self, key: HashPoint[KVKeyType], value: HashPoint[KVValueType]):
        assert isinstance(key, HashPoint)
        assert isinstance(value, HashPoint)
        super().__init__(key)
        self.value = value

    def points(self) -> Iterable[HashPoint]:
        return [self.key, self.value]

    def __bytes__(self):
        return bytes(self.key) + bytes(self.value)

    def __factory__(self) -> RainbowFactory['KeyValue[KVKeyType, KVValueType]']:
        return KeyValueFactory(self.key.factory, self.value.factory)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        key_str, value_str = await gather(
            hash_point_format(self.key, tab),
            hash_point_format(self.value, tab)
        )
        assert isinstance(key_str, str)
        assert isinstance(value_str, str)
        return f'{key_str}' \
               f'{tabulate(tab)}{value_str}'


class KeyValueFactory(
    RainbowFactory[KeyValue[KVKeyType, KVValueType]],
    Generic[KVKeyType, KVValueType]
):
    def __init__(self, key_factory: RainbowFactory[KVKeyType], value_factory: RainbowFactory[KVValueType]):
        assert isinstance(key_factory, RainbowFactory)
        assert isinstance(value_factory, RainbowFactory)
        self.key_factory = key_factory
        self.value_factory = value_factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> KeyValue[KVKeyType, KVValueType]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return KeyValue(
            ResolverOrigin(self.key_factory, source[:HashPoint.HASH_LENGTH], resolver).hash_point(),
            ResolverOrigin(self.value_factory, source[HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    def loose(self) -> RainbowFactory[KeyValue[KVKeyType, KVValueType]]:
        return self
