from typing import Generic, TypeAlias, TypeVar

from rainbowadn.collection.comparison import *
from rainbowadn.collection.trees.binary.actions import *
from rainbowadn.collection.trees.binary.core import *
from rainbowadn.collection.trees.binary.core import BinaryProtocolized
from rainbowadn.core import *
from rainbowadn.flow.core import *

__all__ = ('BinaryReducer', 'VerifySubsetAction', 'CheckResult',)

KeyT = TypeVar('KeyT')
MetadataT = TypeVar('MetadataT')
TreeT = TypeVar('TreeT')
Out = TypeVar('Out')

BP: TypeAlias = BinaryProtocolized[KeyT, MetadataT, TreeT]
PBS: TypeAlias = ProtocolizedBinarySplit[KeyT, MetadataT, TreeT]


class BinaryReducer(
    Reducer[HashPoint[KeyT], Out],
    Generic[Out, KeyT, MetadataT, TreeT]
):
    def __init__(self, protocolized: BP):
        assert isinstance(protocolized, BinaryProtocolized)
        self.protocolized = protocolized

    async def reduce(self, reduce: Reduce[HashPoint[KeyT], Out]) -> Out:
        assert isinstance(reduce, Reduce)
        return await BinaryReducerAction(reduce).on(self.protocolized)

    def loose(self) -> Reducer[HashPoint[KeyT], Out]:
        return self


class BinaryReducerAction(
    BinaryAction[KeyT, MetadataT, TreeT, Out],
    Generic[KeyT, MetadataT, TreeT, Out],
):
    def __init__(self, reduce: Reduce[KeyT, Out]):
        assert isinstance(reduce, Reduce)
        self.reduce = reduce

    async def on_null(self, protocolized: BP) -> Out:
        assert isinstance(protocolized, BinaryProtocolized)
        return self.reduce.initial

    async def on_split(self, case: PBS) -> Out:
        assert isinstance(case, ProtocolizedBinarySplit)
        left: Out
        key: KeyT
        right: Out
        left, right = await gather(
            self.on(case.protocolizedl()),
            self.on(case.protocolizedr()),
        )

        return await self.reduce.merge(
            await self.reduce.reduce(
                left,
                case.split.key
            ),
            right,
        )


class NotFound:
    def __lt__(self, other):
        if isinstance(other, NotFound):
            return False
        elif isinstance(other, Found):
            return True
        elif isinstance(other, SubTree):
            return True
        else:
            return NotImplemented


class Found:
    def __lt__(self, other):
        if isinstance(other, NotFound):
            return False
        elif isinstance(other, Found):
            return False
        elif isinstance(other, SubTree):
            return True
        else:
            return NotImplemented


class SubTree:
    def __lt__(self, other):
        if isinstance(other, NotFound):
            return False
        elif isinstance(other, Found):
            return False
        elif isinstance(other, SubTree):
            return False
        else:
            return NotImplemented


CheckResult: TypeAlias = NotFound | Found | SubTree


class SubCheck(
    CompareAction[KeyT, MetadataT, TreeT, CheckResult],
    Generic[KeyT, MetadataT, TreeT],
):
    def __init__(self, split: PBS):
        assert isinstance(split, ProtocolizedBinarySplit)
        self.split = split
        super().__init__(split.split.key)

    async def on_equal(self, case: PBS, equal: Equal) -> CheckResult:
        assert isinstance(case, ProtocolizedBinarySplit)
        assert isinstance(equal, Equal)
        if case.tree == self.split.tree:
            return SubTree()
        else:
            return Found()

    async def on_left(self, case: PBS) -> CheckResult:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await self.on(case.protocolizedl())

    async def on_right(self, case: PBS) -> CheckResult:
        assert isinstance(case, ProtocolizedBinarySplit)
        return await self.on(case.protocolizedr())

    async def on_null(self, protocolized: BP) -> CheckResult:
        assert isinstance(protocolized, BinaryProtocolized)
        return NotFound()


class SubCheckMapper(
    Mapper[BP, CheckResult],
    Generic[KeyT, MetadataT, TreeT],
):
    def __init__(self, split: PBS):
        assert isinstance(split, ProtocolizedBinarySplit)
        self.split = split

    async def map(self, element: BP) -> CheckResult:
        assert isinstance(element, BinaryProtocolized)
        return await SubCheck(self.split).on(element)

    def loose(self) -> Mapper[BP, CheckResult]:
        return self


class VerifySubsetAction(
    BinaryAction[KeyT, MetadataT, TreeT, bool],
    Generic[KeyT, MetadataT, TreeT],
):
    def __init__(self, reducer: Reducer[BP, CheckResult]):
        assert isinstance(reducer, Reducer)
        self.reducer = reducer

    async def on_null(self, protocolized: BP) -> bool:
        assert isinstance(protocolized, BinaryProtocolized)
        return True

    async def on_split(self, case: PBS) -> bool:
        assert isinstance(case, ProtocolizedBinarySplit)
        reduce: Reduce[BP, CheckResult] = MapReduce(
            SubCheckMapper(case).loose(), VerifySubsetReduce(NotFound())
        ).loose()
        assert isinstance(reduce, Reduce)
        result: CheckResult = await self.reducer.reduce(reduce)
        match result:
            case SubTree():
                return True
            case Found():
                assert_trues(
                    await gather(
                        self.on(case.protocolizedl()),
                        self.on(case.protocolizedr()),
                    )
                )
                return True
            case NotFound():
                raise ValueError('subset check failed')
            case _:
                raise TypeError


class VerifySubsetReduce(
    PureReduce[CheckResult]
):
    async def merge(self, left: CheckResult, right: CheckResult) -> CheckResult:
        return max(left, right)
