import functools
from contextlib import ExitStack
from typing import Callable, Generic, TypeVar

__all__ = ('Instrumentation',)

IType = TypeVar('IType')


class Instrumentation(Generic[IType]):
    deinstrumentation = {}
    _method: Callable
    _wrap: Callable

    def __init__(self, target, methodname: str):
        assert isinstance(methodname, str)
        assert callable(getattr(target, methodname))
        self.target = target
        self.methodname = methodname

    def instrument(self, method, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self: IType) -> IType:
        self: Instrumentation
        assert not hasattr(self, '_method')
        assert not hasattr(self, '_wrap')
        method = getattr(self.target, self.methodname)
        assert callable(method)
        self._method = method

        @functools.wraps(method)
        def wrap(*args, **kwargs):
            nonlocal method
            while method in self.deinstrumentation:
                self._method = method = self.deinstrumentation.pop(method)
            return self.instrument(method, *args, **kwargs)

        self._wrap = wrap

        setattr(self.target, self.methodname, wrap)

        return self

    def schedule_deinstrumentation(self):
        self.deinstrumentation[self._wrap] = self._method
        del self._wrap
        del self._method

    def deinstrument(self):
        while (method := getattr(self.target, self.methodname)) in self.deinstrumentation:
            setattr(self.target, self.methodname, self.deinstrumentation.pop(method))

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.schedule_deinstrumentation()
        self.deinstrument()

    def enter(self: IType, stack: ExitStack) -> IType:
        return stack.enter_context(self)
